from abc import abstractmethod
from typing import List

from bayserver_core.protocol.protocol_exception import ProtocolException
from bayserver_docker_ajp.ajp_command_handler import AjpCommandHandler


class AjpHandler(AjpCommandHandler):

    # Send protocol error to client
    @abstractmethod
    def on_protocol_error(self, e:ProtocolException, stk: List[str]) -> bool:
        pass


from abc import ABCMeta, abstractmethod

from bayserver_core.protocol.command_handler import CommandHandler

class AjpCommandHandler(CommandHandler, metaclass=ABCMeta):
    @abstractmethod
    def handle_forward_request(self, cmd):
        pass

    @abstractmethod
    def handle_data(self, cmd):
        pass

    @abstractmethod
    def handle_send_body_chunk(self, cmd):
        pass

    @abstractmethod
    def handle_send_headers(self, cmd):
        pass

    @abstractmethod
    def handle_shutdown(self, cmd):
        pass

    @abstractmethod
    def handle_end_response(self, cmd):
        pass

    @abstractmethod
    def handle_get_body_chunk(self, cmd):
        pass

    @abstractmethod
    def need_data(self):
        pass



from bayserver_core.bay_log import BayLog

from bayserver_docker_ajp.ajp_command import AjpCommand
from bayserver_docker_ajp.ajp_type import AjpType
from bayserver_core.protocol.protocol_exception import ProtocolException
from bayserver_core.util.http_status import HttpStatus

#
#  Send headers format
#
#  AJP13_SEND_HEADERS :=
#    prefix_code       4
#    http_status_code  (integer)
#    http_status_msg   (string)
#    num_headers       (integer)
#    response_headers *(res_header_name header_value)
#
#  res_header_name :=
#      sc_res_header_name | (string)   [see below for how this is parsed]
#
#  sc_res_header_name := 0xA0 (byte)
#
#  header_value := (string)
#

class CmdSendHeaders(AjpCommand):
    well_known_header_map = {}

    well_known_header_map = {
        "content-type": 0xA001,
        "content-language": 0xA002,
        "content-length": 0xA003,
        "date": 0xA004,
        "last-modified": 0xA005,
        "location": 0xA006,
        "set-cookie": 0xA007,
        "set-cookie2": 0xA008,
        "servlet-engine": 0xA009,
        "status": 0xA00A,
        "www-authenticate": 0xA00B,
    }

    @classmethod
    def get_well_known_header_name(cls, code):
        for name in cls.well_known_header_map.keys():
            if cls.well_known_header_map[name] == code:
                return name

        return None

    def __init__(self):
        super().__init__(AjpType.SEND_HEADERS, False)
        self.headers = {}
        self.status = HttpStatus.OK
        self.desc = None

    def __str__(self):
        return f"SendHeaders: s={self.status} d={self.desc} h={self.headers}"

    def pack(self, pkt):
        acc = pkt.new_ajp_data_accessor()
        acc.put_byte(self.type)
        acc.put_short(self.status)
        acc.put_string(HttpStatus.description(self.status))

        count = 0
        for key in self.headers.keys():
            count += len(self.headers[key])

        acc.put_short(count)

        for name in self.headers.keys():
            code = CmdSendHeaders.well_known_header_map.get(name)

            for value in self.headers[name]:
                if code is not None:
                    acc.put_short(code)
                else:
                    acc.put_string(name)

                acc.put_string(value)


        #  must be called from last line
        super().pack(pkt)

    def unpack(self, pkt):
        acc = pkt.new_ajp_data_accessor()
        prefix_code = acc.get_byte()
        if prefix_code != AjpType.SEND_HEADERS:
            raise ProtocolException("Expected SEND_HEADERS")

        self.status = acc.get_short()
        self.desc = acc.get_string()
        count = acc.get_short()
        for i in range(count):
            code = acc.get_short()
            name = CmdSendHeaders.get_well_known_header_name(code)
            if name is None:
                name = acc.get_string_by_len(code)

            value = acc.get_string()
            self.add_header(name, value)
        #BayLog.info("%s", self)

    def handle(self, handler):
        return handler.handle_send_headers(self)

    def get_header(self, name):
        values = self.headers.get(name.lower())
        if values is None or len(values) == 0:
            return None
        else:
            return values[0]

    def add_header(self, name, value):
        values = self.headers.get(name)
        if values is None:
            values = []
            self.headers[name] = values

        values.append(value)




from bayserver_docker_ajp.ajp_command import AjpCommand
from bayserver_docker_ajp.ajp_type import AjpType

#
#  Shutdown command format
#
#    none
#

class CmdShutdown(AjpCommand):

    def __init__(self):
        super(AjpType.SHUTDOWN, True)

    def unpack(self, pkt):
        super().unpack(pkt)

    def pack(self, pkt):
        super().pack(pkt)

    def handle(self, handler):
        return handler.handle_shutdown(self)



from bayserver_core.protocol.packet_factory import PacketFactory
from bayserver_docker_ajp.ajp_packet import AjpPacket


class AjpPacketFactory(PacketFactory):

    def create_packet(self, typ):
        return AjpPacket(typ)



from bayserver_core.protocol.protocol_handler import ProtocolHandler
from bayserver_core.protocol.packet_packer import PacketPacker
from bayserver_core.protocol.command_packer import CommandPacker

from bayserver_docker_ajp.ajp_command_unpacker import AjpCommandUnPacker
from bayserver_docker_ajp.ajp_docker import AjpDocker
from bayserver_docker_ajp.ajp_handler import AjpHandler
from bayserver_docker_ajp.ajp_packet_unpacker import AjpPacketUnPacker
from bayserver_docker_ajp.ajp_command_handler import AjpCommandHandler
from bayserver_docker_ajp.command.cmd_data import CmdData
from bayserver_docker_ajp.command.cmd_send_body_chunk import CmdSendBodyChunk


class AjpProtocolHandler(ProtocolHandler):

    def __init__(self,
                 h1_handler: AjpHandler,
                 packet_unpacker: AjpPacketUnPacker,
                 packet_packer: PacketPacker,
                 command_unpacker: AjpCommandUnPacker,
                 command_packer: CommandPacker,
                 svr_mode: bool):
        super().__init__(
            packet_unpacker,
            packet_packer,
            command_unpacker,
            command_packer,
            h1_handler,
            svr_mode
        )

    def __str__(self):
        return f"pch[{self.ship}]"


    ######################################################
    # Implements ProtocolHandler
    ######################################################

    def protocol(self):
        return AjpDocker.PROTO_NAME

    def max_req_packet_data_size(self):
        return CmdData.MAX_DATA_LEN

    def max_res_packet_data_size(self):
        return CmdSendBodyChunk.MAX_CHUNKLEN




#
# Ajp packet type
#

class AjpType:
    DATA = 0
    FORWARD_REQUEST = 2
    SEND_BODY_CHUNK = 3
    SEND_HEADERS = 4
    END_RESPONSE = 5
    GET_BODY_CHUNK = 6
    SHUTDOWN = 7
    PING = 8
    CPING = 10


from typing import Any

from bayserver_core.bay_log import BayLog
from bayserver_core.protocol.packet_unpacker import PacketUnPacker

from bayserver_core.agent.next_socket_action import NextSocketAction
from bayserver_core.util.simple_buffer import SimpleBuffer
from bayserver_core.util.char_util import CharUtil

from bayserver_docker_ajp.ajp_packet import AjpPacket
from bayserver_docker_ajp.ajp_type import AjpType

class AjpPacketUnPacker(PacketUnPacker):
    STATE_READ_PREAMBLE = 1
    STATE_READ_BODY = 2
    STATE_END = 3

    def __init__(self, pkt_store, cmd_unpacker):

        self.preamble_buf = None
        self.body_buf = None
    
        self.state = None
    
        self.body_len = None
        self.read_bytes = None
        self.type = None
        self.to_server = None
        self.need_data = None
        self.pkt_store = pkt_store
        self.cmd_unpacker = cmd_unpacker
        self.preamble_buf = SimpleBuffer()
        self.body_buf = SimpleBuffer()
        self.reset()

    def reset(self):
        self.state = AjpPacketUnPacker.STATE_READ_PREAMBLE
        self.body_len = 0
        self.read_bytes = 0
        self.need_data = False
        self.preamble_buf.reset()
        self.body_buf.reset()

    def bytes_received(self, buf: bytes, adr: Any):
        suspend = False
        pos = 0

        while pos < len(buf):
            if self.state == AjpPacketUnPacker.STATE_READ_PREAMBLE:
                length = AjpPacket.PREAMBLE_SIZE - len(self.preamble_buf)
                if len(buf) - pos < length:
                    length = len(buf) - pos

                self.preamble_buf.put(buf, pos, length)
                pos += length

                if len(self.preamble_buf) == AjpPacket.PREAMBLE_SIZE:
                    self.preamble_read()
                    self.change_state(AjpPacketUnPacker.STATE_READ_BODY)

            if self.state == AjpPacketUnPacker.STATE_READ_BODY:
                length = self.body_len - len(self.body_buf)
                if length > len(buf) - pos:
                    length = len(buf) - pos

                self.body_buf.put(buf, pos, length)
                pos += length

                if len(self.body_buf) == self.body_len:
                    self.body_read()
                    self.change_state(AjpPacketUnPacker.STATE_END)

            if self.state == AjpPacketUnPacker.STATE_END:

                pkt = self.pkt_store.rent(self.type)
                pkt.to_server = self.to_server
                pkt.new_ajp_header_accessor().put_bytes(self.preamble_buf.buf, 0, len(self.preamble_buf))
                pkt.new_ajp_data_accessor().put_bytes(self.body_buf.buf, 0, len(self.body_buf))

                try:
                    next_action = self.cmd_unpacker.packet_received(pkt)
                finally:
                    self.pkt_store.Return(pkt)

                self.reset()
                self.need_data = self.cmd_unpacker.need_data()

                if next_action == NextSocketAction.SUSPEND:
                    suspend = True
                elif next_action != NextSocketAction.CONTINUE:
                    return next_action

        BayLog.debug("ajp next read")
        if suspend:
            return NextSocketAction.SUSPEND
        else:
            return NextSocketAction.CONTINUE

    def change_state(self, new_state):
        self.state = new_state

    def preamble_read(self):
        data = self.preamble_buf.buf

        if data[0] == 0x12 and data[1] == 0x34:
            self.to_server = True
        elif data[0] == CharUtil.A_BYTE and data[1] == CharUtil.B_BYTE:
            self.to_server = False

        else:
            raise RuntimeError("Must be start with 0x1234 or 'AB'")


        self.body_len = ((data[2] << 8) | (data[3] & 0xff)) & 0xffff
        BayLog.trace("ajp: read packet preamble: bodyLen=%d", self.body_len)

    def body_read(self):
        if self.need_data:
            self.type = AjpType.DATA
        else:
            self.type = self.body_buf.buf[0] & 0xff


from bayserver_docker_ajp.ajp_command import AjpCommand
from bayserver_docker_ajp.ajp_packet import AjpPacket
from bayserver_docker_ajp.ajp_type import AjpType

#
#  Send body chunk format
#
#  AJP13_SEND_BODY_CHUNK :=
#    prefix_code   (byte) 0x03
#    chunk_length  (integer)
#    chunk         *(byte)
#

class CmdSendBodyChunk(AjpCommand):

    MAX_CHUNKLEN = AjpPacket.MAX_DATA_LEN - 4

    def __init__(self, buf, ofs, length):
        super().__init__(AjpType.SEND_BODY_CHUNK, False)
        self.chunk = buf
        self.offset = ofs
        self.length = length

    def pack(self, pkt):
        if self.length > CmdSendBodyChunk.MAX_CHUNKLEN:
            raise RuntimeError("IllegalArgument")

        acc = pkt.new_ajp_data_accessor()
        acc.put_byte(self.type)
        acc.put_short(self.length)
        acc.put_bytes(self.chunk, self.offset, self.length)
        acc.put_byte(0)  # maybe document bug

        #  must be called from last line
        super().pack(pkt)

    def unpack(self, pkt):
        acc = pkt.new_ajp_data_accessor()
        acc.get_byte()  # code
        self.length = acc.get_short()
        if self.chunk is None or self.length > len(self.chunk):
            self.chunk = bytearray(self.length)

        acc.get_bytes(self.chunk, 0, self.length)

    def handle(self, handler):
        return handler.handle_send_body_chunk(self)


from bayserver_core.bay_log import BayLog
from bayserver_docker_ajp.ajp_command import AjpCommand
from bayserver_docker_ajp.ajp_type import AjpType

from bayserver_core.protocol.protocol_exception import ProtocolException
from bayserver_core.util.headers import Headers

#
#  AJP protocol
#     https://tomcat.apache.org/connectors-doc/ajp/ajpv13a.html
#
#  AJP13_FORWARD_REQUEST :=
#      prefix_code      (byte) 0x02 = JK_AJP13_FORWARD_REQUEST
#      method           (byte)
#      protocol         (string)
#      req_uri          (string)
#      remote_addr      (string)
#      remote_host      (string)
#      server_name      (string)
#      server_port      (integer)
#      is_ssl           (boolean)
#      num_headers      (integer)
#      request_headers(req_header_name req_header_value)
#      attributes     (attribut_name attribute_value)
#      request_terminator (byte) OxFF
#

class CmdForwardRequest(AjpCommand):
    method_map = None
    well_known_header_map = None
    attribute_name_map = None

    method_map = {
        1: "OPTIONS",
        2: "GET",
        3: "HEAD",
        4: "POST",
        5: "PUT",
        6: "DELETE",
        7: "TRACE",
        8: "PROPFIND",
        9: "PROPPATCH",
        10: "MKCOL",
        11: "COPY",
        12: "MOVE",
        13: "LOCK",
        14: "UNLOCK",
        15: "ACL",
        16: "REPORT",
        17: "VERSION_CONTROL",
        18: "CHECKIN",
        19: "CHECKOUT",
        20: "UNCHECKOUT",
        21: "SEARCH",
        22: "MKWORKSPACE",
        23: "UPDATE",
        24: "LABEL",
        25: "MERGE",
        26: "BASELINE_CONTROL",
        27: "MKACTIVITY",
    }

    @classmethod
    def get_method_code(cls, method):
        for key in CmdForwardRequest.method_map.keys():
            if cls.method_map[key].lower() == method.lower():
                return key


        return -1

    well_known_header_map = {
        0xA001: "Accept",
        0xA002: "Accept-Charset",
        0xA003: "Accept-Encoding",
        0xA004: "Accept-Language",
        0xA005: "Authorization",
        0xA006: "Connection",
        0xA007: "Content-Type",
        0xA008: "Content-Length",
        0xA009: "Cookie",
        0xA00A: "Cookie2",
        0xA00B: "Host",
        0xA00C: "Pragma",
        0xA00D: "Referer",
        0xA00E: "User-Agent",
    }

    @classmethod
    def get_well_known_header_code(cls, name):
        for key in cls.well_known_header_map.keys():
            if cls.well_known_header_map[key].lower() == name.lower():
                return key

        return -1

    attribute_name_map = {
        0x01: "?context",
        0x02: "?servlet_path",
        0x03: "?remote_user",
        0x04: "?auth_type",
        0x05: "?query_string",
        0x06: "?route",
        0x07: "?ssl_cert",
        0x08: "?ssl_cipher",
        0x09: "?ssl_session",
        0x0A: "?req_attribute",
        0x0B: "?ssl_key_size",
        0x0C: "?secret",
        0x0D: "?stored_method",
    }

    @classmethod
    def get_attribute_code(cls, atr):
        for key in cls.attribute_name_map.keys():
            if cls.attribute_name_map[key].lower() == atr.lower():
                return key

        return -1



    def __init__(self):
        super().__init__(AjpType.FORWARD_REQUEST, True)
        self.method = None
        self.protocol = None
        self.req_uri = None
        self.remote_addr = None
        self.remote_host = None
        self.server_name = None
        self.server_port = None
        self.is_ssl = None
        self.headers = Headers()
        self.attributes = {}


    def __str__(self):
        return f"ForwardRequest(m={self.method} p={self.protocol} u={self.req_uri} ra={self.remote_addr} rh={self.remote_host} sn={self.server_name} sp={self.server_port} ss={self.is_ssl} h={self.headers}"


    def pack(self, pkt):
        #BayLog.info("%s", self)
        acc = pkt.new_ajp_data_accessor()
        acc.put_byte(self.type)  # prefix code
        acc.put_byte(CmdForwardRequest.get_method_code(self.method))
        acc.put_string(self.protocol)
        acc.put_string(self.req_uri)
        acc.put_string(self.remote_addr)
        acc.put_string(self.remote_host)
        acc.put_string(self.server_name)
        acc.put_short(self.server_port)
        acc.put_byte(1 if self.is_ssl else 0)
        self.write_request_headers(acc)
        self.write_attributes(acc)

        #  must be called from last line
        super().pack(pkt)

    def unpack(self, pkt):
        super().unpack(pkt)
        acc = pkt.new_ajp_data_accessor()
        acc.get_byte()  # prefix code
        self.method = CmdForwardRequest.method_map.get(acc.get_byte())
        self.protocol = acc.get_string()
        self.req_uri = acc.get_string()
        self.remote_addr = acc.get_string()
        self.remote_host = acc.get_string()
        self.server_name = acc.get_string()
        self.server_port = acc.get_short()
        self.is_ssl = acc.get_byte() == 1

        self.read_request_headers(acc)
        self.read_attributes(acc)

    def handle(self, handler):
        return handler.handle_forward_request(self)


    #
    # private
    #
    def read_request_headers(self, acc):
        count = acc.get_short()
        for i in range(count):
            code = acc.get_short()

            if code >= 0xA000:
                name = CmdForwardRequest.well_known_header_map.get(code)

                if name is None:
                    raise ProtocolException("Invalid header")

            else:
                # code is length of header name
                name = acc.get_string_by_len(code)

            value = acc.get_string()
            self.headers.add(name, value)
            # BayLog.trace "ForwardRequest header: #{name}=#{value}"

    def read_attributes(self, acc):
        while True:
            code = acc.get_byte()

            if code == 0xFF:
                break
            elif code == 0x0A:
                name = acc.get_string()
            else:
                name = CmdForwardRequest.attribute_name_map[code]
            if name is None:
                raise ProtocolException(f"Invalid attribute: code={code}")

            if code == 0x0B:  # "?ssl_key_size"
                value = acc.get_short()
                self.attributes[name] = str(value)
            else:
                value = acc.get_string()
                self.attributes[name] = value

    def write_request_headers(self, acc):
        h_list = []
        for name in self.headers.names():
            for value in self.headers.values(name):
                h_list.append([name, value])


        acc.put_short(len(h_list))
        for item in h_list:
            code = CmdForwardRequest.get_well_known_header_code(item[0])
            if code != -1:
                acc.put_short(code)
            else:
                acc.put_string(item[0])

            acc.put_string(item[1])

    def write_attributes(self, acc):
        for name in self.attributes.keys():
            value = self.attributes[name]
            code = CmdForwardRequest.get_attribute_code(name)
            if code != -1:
                acc.put_byte(code)
            else:
                acc.put_string(name)

            acc.put_string(value)

        acc.put_byte(0xFF)  # terminator code

